﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using BeanboxSoftware.BeanMap;

namespace BeanboxSoftware.BeanMapTest.Generics
{
  [TestClass]
  public class MapKVTest
  {
    [TestMethod]
    public void BasicUseTest()
    {
      var map = new Map<int, string>();

      map[1] = "jun";
      map[2] = "cuib";

      Assert.IsFalse(map.ContainsKey(3), "ContainsKey(3)");

      map.DefaultValue = "mac'a";

      Assert.AreEqual("mac'a", map[3], "map[3] with default error value");
      Assert.IsFalse(map.ContainsKey(3), "ContainsKey(3) after default error value");

      map.DefaultGeneration = key => key.ToString();

      Assert.AreEqual("3", map[3], "map[3] with generator");
      Assert.IsTrue(map.ContainsKey(3), "ContainsKey(3) after generator");
    }

    [TestMethod]
    public void PropertiesTest()
    {
      var map = new Map<int, string>();
      map[0] = "mac'a";
      map[1] = "jun";

      Assert.AreEqual(2, map.Count, "count");

      Assert.IsTrue(map.Values.Contains("mac'a"), "values should contain mac'a");
      Assert.IsTrue(map.Values.Contains("jun"), "values should contain jun");
      Assert.IsFalse(map.Values.Contains("cuib"), "values should not contain cuib");

      Assert.IsTrue(map.Keys.Contains(0), "keys should contain 0");
      Assert.IsTrue(map.Keys.Contains(1), "keys should contain 1");
      Assert.IsFalse(map.Keys.Contains(2), "keys should contain 2");
    }

    [TestMethod]
    public void ListCtorTest()
    {
      var list = new List<int>();
      list.Add(1);
      list.Add(2);

      var map = new Map<string, int>(list, item => item.ToString());

      Assert.AreEqual(1, map["1"], "map[1]");
      Assert.AreEqual(2, map["2"], "map[2]");

      map.KeyNotFoundBehavior = KeyNotFoundBehavior.ThrowKeyNotFoundException;

      TestUtils.AssertExceptionThrown<KeyNotFoundException>(
        () =>
        {
          Assert.Fail(map["3"].ToString());
        }, "3 doesn't exist and should throw");
    }

    [TestMethod]
    public void KeyValueSetBaseTest()
    {
      var map = new Map<int, string>();
      map[3] = "three";

      var kvs = map.KeyValues;

      Assert.AreEqual("three", kvs.First().Value, "kv value 0");
      Assert.AreEqual(3, kvs.First().Key, "kv key 0");
    }

    [TestMethod]
    public void KeyNotFoundBehaviorTest()
    {
      var map = new Map<int, int>();
      map.DefaultGeneration = (key) => { return -1; };

      Assert.AreEqual(KeyNotFoundBehavior.DefaultGeneration, map.KeyNotFoundBehavior,
        "behavior after setting generator");

      map.DefaultValue = 41;

      Assert.AreEqual(41, map[414], "default value");
      Assert.IsFalse(map.ContainsKey(414), "default value should not create a key");
    }

    [TestMethod]
    public void RemoveTest()
    {
      var map = new Map<int, int>();
      map[1] = 2;

      Assert.IsFalse(map.Remove(2), "2 not found");
      Assert.IsTrue(map.Remove(1), "1 not found");
    }

    [TestMethod]
    public void ClearTest()
    {
      var map = new Map<int, int>();
      map[1] = 2;

      map.Clear();

      Assert.AreEqual(0, map.Count, "map count after clear");
    }

    [TestMethod]
    public void ContainsKeyTest()
    {
      var map = new Map<int, string>();
      map[1] = "jun";

      Assert.IsTrue(map.ContainsKey(1), "contains 1");
      Assert.IsFalse(map.ContainsKey(2), "contains 2");
    }

    [TestMethod]
    public void DictionaryLikeApiTest()
    {
      var dict = new Dictionary<int, string>()
      {
        {1, "one"},
        {2, "two"},
        {3, "three"},
      };

      var map = new Map<int, string>()
      {
        {1, "one"},
        {2, "two"},
        {3, "three"},
      };

      foreach (var key in map.Keys)
      {
        Assert.AreEqual(dict[key], map[key], "dict style init list for key: " + key.ToString());
      }

      TestUtils.AssertExceptionThrown<ArgumentException>(() => { map.Add(1, "other"); }, 
        "add should act like dict act");

      foreach (var it in map)
      {
        var kv = (KeyValuePair<int, string>)it;

        Assert.AreEqual(map[kv.Key], kv.Value, "kv key: " + kv.Key.ToString());
      }
    }

    [TestMethod]
    public void CopyCtor()
    {
      var map = new Map<int, string>();
      map[10] = "10";
      map[15] = "15";

      var copy = new Map<int, string>(map);

      Assert.AreEqual(map[10], copy[10], "10");
      Assert.AreEqual(map[15], copy[15], "15");

      copy[20] = "20";
      map.Remove(10);

      Assert.IsFalse(map.ContainsKey(20), "20");
      Assert.IsTrue(copy.ContainsKey(10), "10");
    }

    class Thing
    {
      public int Foo { get; set; }
      public int Bar { get; set; }
      public string Baz { get; set; }
    }

    [TestMethod]
    public void ListToMapTest()
    {
      var things = new List<Thing>
                 {
                     new Thing {Foo = 3, Bar = 4, Baz = "brown"},
                     new Thing {Foo = 6, Bar = 4, Baz = "jumps"}
                 };

      var thingMap = Map<int, string>.From(things, t => t.Foo, t => t.Baz);

      Assert.IsTrue(thingMap.ContainsKey(3));
      Assert.AreEqual("brown", thingMap[3]);

      thingMap.DefaultValue = string.Empty;
      Assert.AreEqual("brown", thingMap[3]);
      Assert.AreEqual(string.Empty, thingMap[2]);
    }
  }
}
